"""Red-black tree, where keys are stored as start/end timestamps.
This is a basic interval tree that holds half-open intervals:
  [start, end)
Intervals must not overlap.  Fixing that would involve making this
into an augmented interval tree as described in CLRS 14.3.
"""

import sys

class RBNode(object):
    """One node of the Red/Black tree, containing a key (start, end)
    and value (obj)"""
    def __init__(self, start, end, obj = None):
        self.obj = obj
        self.start = start
        self.end = end
        self.red = False
        self.left = None
        self.right = None

    def __str__(self):
        if self.red:
            color = "R"
        else:
            color = "B"
        return ("[node "
                + str(obj) + " "
                + str(self.start) + " -> " + str(self.end) + " "
                + color + "]")

class RBTree(object):
    """Red/Black tree"""

    # Init
    def __init__(self):
        self.nil = RBNode(start = sys.float_info.min,
                          end = sys.float_info.min)
        self.nil.left = self.nil
        self.nil.right = self.nil
        self.nil.parent = self.nil
        self.nil.nil = True

        self.root = RBNode(start = sys.float_info.max,
                           end = sys.float_info.max)
        self.root.left = self.nil
        self.root.right = self.nil
        self.root.parent = self.nil

    # Rotations and basic operations
    def __rotate_left(self, x):
        """Rotate left:
        #   x           y
        #  / \   -->   / \
        # z   y       x   w
        #    / \     / \
        #   v   w   z   v
        """
        y = x.right
        x.right = y.left
        if y.left is not self.nil:
            y.left.parent = x
        y.parent = x.parent
        if x is x.parent.left:
            x.parent.left = y
        else:
            x.parent.right = y
        y.left = x
        x.parent = y

    def __rotate_right(self, y):
        """Rotate right:
        #     y           x
        #    / \   -->   / \
        #   x   w       z   y
        #  / \             / \
        # z   v           v   w
        """
        x = y.left
        y.left = x.right
        if x.right is not self.nil:
            x.right.parent = y
        x.parent = y.parent
        if y is y.parent.left:
            y.parent.left = x
        else:
            y.parent.right = x
        x.right = y
        y.parent = x

    def __successor(self, x):
        """Returns the successor of x"""
        y = x.right
        if y is not self.nil:
            while y.left is not self.nil:
                y = y.left
        else:
            y = x.parent
            while x is y.right:
                x = y
                y = y.parent
            if y is self.root:
                return self.nil
        return y

    def __predecessor(self, x):
        """Returns the predecessor of RBNode x"""
        y = x.left
        if y is not self.nil:
            while y.right is not self.nil:
                y = y.right
        else:
            y = x.parent
            while x is y.left:
                if y is self.root:
                    y = self.nil
                    break
                x = y
                y = y.parent
        return y

    # Insertion
    def insert(self, z):
        """Insert RBNode z into RBTree and rebalance as necessary"""
        z.left = self.nil
        z.right = self.nil
        y = self.root
        x = self.root.left
        while x is not self.nil:
            y = x
            if (x.start > z.start or (x.start == z.start and x.end > z.end)):
                x = x.left
            else:
                x = x.right
        z.parent = y
        if (y is self.root or
            (y.start > z.start or (y.start == z.start and y.end > z.end))):
            y.left = z
        else:
            y.right = z
        # relabel/rebalance
        self.__insert_fixup(z)

    def __insert_fixup(self, x):
        """Rebalance/fix RBTree after a simple insertion of RBNode x"""
        x.red = True
        while x.parent.red:
            if x.parent is x.parent.parent.left:
                y = x.parent.parent.right
                if y.red:
                    x.parent.red = False
                    y.red = False
                    x.parent.parent.red = True
                    x = x.parent.parent
                else:
                    if x is x.parent.right:
                        x = x.parent
                        self.__rotate_left(x)
                    x.parent.red = False
                    x.parent.parent.red = True
                    self.__rotate_right(x.parent.parent)
            else: # same as above, left/right switched
                y = x.parent.parent.left
                if y.red:
                    x.parent.red = False
                    y.red = False
                    x.parent.parent.red = True
                    x = x.parent.parent
                else:
                    if x is x.parent.left:
                        x = x.parent
                        self.__rotate_right(x)
                    x.parent.red = False
                    x.parent.parent.red = True
                    self.__rotate_left(x.parent.parent)
        self.root.left.red = False

    # Deletion
    def delete(self, z):
        if z.left is None or z.right is None:
            raise AttributeError("you can only delete a node object "
                                 + "from the tree; use find() to get one")
        if z.left is self.nil or z.right is self.nil:
            y = z
        else:
            y = self.__successor(z)
        if y.left is self.nil:
            x = y.right
        else:
            x = y.left
        x.parent = y.parent
        if x.parent is self.root:
            self.root.left = x
        else:
            if y is y.parent.left:
                y.parent.left = x
            else:
                y.parent.right = x
        if y is not z:
            # y is the node to splice out, x is its child
            y.left = z.left
            y.right = z.right
            y.parent = z.parent
            z.left.parent = y
            z.right.parent = y
            if z is z.parent.left:
                z.parent.left = y
            else:
                z.parent.right = y
            if not y.red:
                y.red = z.red
                self.__delete_fixup(x)
            else:
                y.red = z.red
        else:
            if not y.red:
                self.__delete_fixup(x)

    def __delete_fixup(self, x):
        """Rebalance/fix RBTree after a deletion.  RBNode x is the
        child of the spliced out node."""
        rootLeft = self.root.left
        while not x.red and x is not rootLeft:
            if x is x.parent.left:
                w = x.parent.right
                if w.red:
                    w.red = False
                    x.parent.red = True
                    self.__rotate_left(x.parent)
                    w = x.parent.right
                if not w.right.red and not w.left.red:
                    w.red = True
                    x = x.parent
                else:
                    if not w.right.red:
                        w.left.red = False
                        w.red = True
                        self.__rotate_right(w)
                        w = x.parent.right
                    w.red = x.parent.red
                    x.parent.red = False
                    w.right.red = False
                    self.__rotate_left(x.parent)
                    x = rootLeft # exit loop
            else: # same as above, left/right switched
                w = x.parent.left
                if w.red:
                    w.red = False
                    x.parent.red = True
                    self.__rotate_right(x.parent)
                    w = x.parent.left
                if not w.left.red and not w.right.red:
                    w.red = True
                    x = x.parent
                else:
                    if not w.left.red:
                        w.right.red = False
                        w.red = True
                        self.__rotate_left(w)
                        w = x.parent.left
                    w.red = x.parent.red
                    x.parent.red = False
                    w.left.red = False
                    self.__rotate_right(x.parent)
                    x = rootLeft # exit loop
        x.red = False

    # Rendering
    def __render_dot_node(self, node, max_depth = 20):
        from printf import sprintf
        """Render a single node and its children into a dot graph fragment"""
        if max_depth == 0:
            return ""
        if node is self.nil:
            return ""
        def c(red):
            if red:
                return 'color="#ff0000", style=filled, fillcolor="#ffc0c0"'
            else:
                return 'color="#000000", style=filled, fillcolor="#c0c0c0"'
        s = sprintf("%d [label=\"%g\\n%g\", %s];\n",
                    id(node),
                    node.start, node.end,
                    c(node.red))

        if node.left is self.nil:
            s += sprintf("L%d [label=\"-\", %s];\n", id(node), c(False))
            s += sprintf("%d -> L%d [label=L];\n", id(node), id(node))
        else:
            s += sprintf("%d -> %d [label=L];\n", id(node), id(node.left))
        if node.right is self.nil:
            s += sprintf("R%d [label=\"-\", %s];\n", id(node), c(False))
            s += sprintf("%d -> R%d [label=R];\n", id(node), id(node))
        else:
            s += sprintf("%d -> %d [label=R];\n", id(node), id(node.right))
        s += self.__render_dot_node(node.left, max_depth-1)
        s += self.__render_dot_node(node.right, max_depth-1)
        return s

    def render_dot(self, title = "RBTree"):
        """Render the entire RBTree as a dot graph"""
        return ("digraph rbtree {\n"
                + self.__render_dot_node(self.root.left)
                + "}\n");

    def render_dot_live(self, title = "RBTree"):
        """Render the entire RBTree as a dot graph, live GTK view"""
        import gtk
        import gtk.gdk
        sys.path.append("/usr/share/xdot")
        import xdot
        xdot.Pen.highlighted = lambda pen: pen
        s = ("digraph rbtree {\n"
             + self.__render_dot_node(self.root)
             + "}\n");
        window = xdot.DotWindow()
        window.set_dotcode(s)
        window.set_title(title + " - any key to close")
        window.connect('destroy', gtk.main_quit)
        def quit(widget, event):
            if not event.is_modifier:
                window.destroy()
                gtk.main_quit()
        window.widget.connect('key-press-event', quit)
        gtk.main()

    # Walking, searching
    def __iter__(self):
        return self.inorder(self.root.left)

    def inorder(self, x = None):
        """Generator that performs an inorder walk for the tree
        starting at RBNode x"""
        if x is None:
            x = self.root.left
        while x.left is not self.nil:
            x = x.left
        while x is not self.nil:
            yield x
            x = self.__successor(x)

    def __find_all(self, start, end, x):
        """Find node with the specified (start,end) key.
        Also returns the largest node less than or equal to key,
        and the smallest node greater or equal to than key."""
        if x is None:
            x = self.root.left
        largest = self.nil
        smallest = self.nil
        while x is not self.nil:
            if start < x.start:
                smallest = x
                x = x.left # start <
            elif start == x.start:
                if end < x.end:
                    smallest = x
                    x = x.left # start =, end <
                elif end == x.end: # found it
                    smallest = x
                    largest = x
                    break
                else:
                    largest = x
                    x = x.right # start =, end >
            else:
                largest = x
                x = x.right # start >
        return (x, smallest, largest)

    def find(self, start, end, x = None):
        """Find node with the key == (start,end), or None"""
        y = self.__find_all(start, end, x)[1]
        return y if y is not self.nil else None

    def find_right(self, start, end, x = None):
        """Find node with the smallest key >= (start,end), or None"""
        y = self.__find_all(start, end, x)[1]
        return y if y is not self.nil else None

    def find_left(self, start, end, x = None):
        """Find node with the largest key <= (start,end), or None"""
        y = self.__find_all(start, end, x)[2]
        return y if y is not self.nil else None

    # Intersections
    def intersect(self, start, end):
        """Generator that returns nodes that overlap the given
        (start,end) range, for the tree rooted at RBNode x.

        NOTE: this assumes non-overlapping intervals."""
        # Start with the leftmost node before the starting point
        n = self.find_left(start, start)
        # If we didn't find one, look for the leftmode node before the
        # ending point instead.
        if n is None:
            n = self.find_left(end, end)
        # If we still didn't find it, there are no intervals that intersect.
        if n is None:
            return

        # Now yield this node and all successors until their endpoints

        if False:
            yield
        return
